@@ -1,8 +1,12 @@
+"""
+
+.. include:: ../README.md
+
+"""
+__pdoc__ = {
+    'Proxy.__': True
+}
 from .view import *
 from .excs import *
 from .core import *
 from .util import *
-
-
-def get(view, **kw):
-    return view.get_(**kw)
